@@ -6,9 +6,11 @@ use warnings;
 use Log::ger;
 
 our $AUTHORITY = 'cpan:PERLANCAR'; # AUTHORITY
-our $DATE = '2022-02-24'; # DATE
+our $DATE = '2023-01-15'; # DATE
 our $DIST = 'App-MetaCPANUtils'; # DIST
-our $VERSION = '0.006'; # VERSION
+our $VERSION = '0.007'; # VERSION
+
+use File::chdir;
 
 our %SPEC;
 
@@ -46,6 +48,8 @@ our $module_fields = [
 
 our %argopt_release_fields = (
     fields => {
+        'x.name.is_plural' => 1,
+        'x.name.singular' => 'field',
         schema => ['array*', of=>['str*', in=>[ map {$_->[0]} @$release_fields ]]],
         default => [ map {$_->[0]} grep {$_->[2]} @$release_fields ],
         cmdline_aliases=>{f=>{}},
@@ -54,6 +58,8 @@ our %argopt_release_fields = (
 );
 our %argopt_distribution_fields = (
     fields => {
+        'x.name.is_plural' => 1,
+        'x.name.singular' => 'field',
         schema => ['array*', of=>['str*', in=>[ map {$_->[0]} @$distribution_fields ]]],
         default => [ map {$_->[0]} grep {$_->[2]} @$distribution_fields ],
         cmdline_aliases=>{f=>{}},
@@ -62,6 +68,8 @@ our %argopt_distribution_fields = (
 );
 our %argopt_module_fields = (
     fields => {
+        'x.name.is_plural' => 1,
+        'x.name.singular' => 'field',
         schema => ['array*', of=>['str*', in=>[ map {$_->[0]} @$module_fields ]]],
         default => [ map {$_->[0]} grep {$_->[2]} @$module_fields ],
         cmdline_aliases=>{f=>{}},
@@ -167,6 +175,7 @@ sub _resultset_to_envres {
             $row->{abstract}     = $obj->abstract     if grep {$_ eq 'abstract'}     @$wanted_fields;
             $row->{first}        = $obj->first        if grep {$_ eq 'first'}        @$wanted_fields;
             $row->{status}       = $obj->status       if grep {$_ eq 'status'}       @$wanted_fields;
+            $row->{download_url} = $obj->download_url if grep {$_ eq 'download_url'} @$wanted_fields;
         } elsif (ref $obj eq 'MetaCPAN::Client::Distribution') {
             $row->{distribution} = $obj->name         if grep {$_ eq 'distribution'} @$wanted_fields;
         } elsif (ref $obj eq 'MetaCPAN::Client::Module') {
@@ -229,6 +238,7 @@ sub list_recent_metacpan_releases {
     my %args = @_;
 
     my $mcpan = MetaCPAN::Client->new;
+    log_trace "[MetaCPAN::Client] Requesting recent %d release(s) ...", $args{n};
     my $recent = $mcpan->recent($args{n});
     _resultset_to_envres($recent, $args{fields});
 }
@@ -282,7 +292,6 @@ sub list_metacpan_releases {
     push @{ $query->{all} }, {distribution=>$args{distribution}} if defined $args{distribution};
     push @{ $query->{all} }, {status=>$args{status}}             if defined $args{status};
     push @{ $query->{all} }, {first=>$args{first}}               if defined $args{first};
-    log_trace "MetaCPAN API query: %s", $query;
 
     my $params = {};
     $params->{_source} = _fields_to_source($args{fields}, $release_fields);
@@ -299,8 +308,8 @@ sub list_metacpan_releases {
         $params->{sort} = [{name=>{order=>'asc'}}]  if $args{sort} eq 'release';
         $params->{sort} = [{name=>{order=>'desc'}}] if $args{sort} eq '-release';
     }
-    log_trace "MetaCPAN API query params: %s", $params;
 
+    log_trace "[MetaCPAN::Client] Requesting releases (query=%s, params=%s) ...", $query, $params;
     my $res = $mcpan->release($query, $params);
 
     _resultset_to_envres($res, $args{fields});
@@ -325,7 +334,6 @@ sub list_metacpan_distributions {
 
     my $query = {all=>[]};
     push @{ $query->{all} }, {author=>$args{author}}             if defined $args{author};
-    log_trace "MetaCPAN API query: %s", $query;
 
     my $params = {};
     $params->{_source} = _fields_to_source($args{fields}, $distribution_fields);
@@ -333,8 +341,8 @@ sub list_metacpan_distributions {
         $params->{sort} = [{name=>{order=>'asc'}}]  if $args{sort} eq 'distribution';
         $params->{sort} = [{name=>{order=>'desc'}}] if $args{sort} eq '-distribution';
     }
-    log_trace "MetaCPAN API query params: %s", $params;
 
+    log_trace "[MetaCPAN::Client] Requesting distributions (query=%s, params=%s) ...", $query, $params;
     my $res = $mcpan->distribution($query, $params);
 
     _resultset_to_envres($res, $args{fields});
@@ -371,7 +379,6 @@ sub list_metacpan_modules {
     my $query = {all=>[]};
     push @{ $query->{all} }, {author=>$args{author}}             if defined $args{author};
     push @{ $query->{all} }, {status=>$args{status}}             if defined $args{status};
-    log_trace "MetaCPAN API query: %s", $query;
 
     my $params = {};
     $params->{_source} = _fields_to_source($args{fields}, $module_fields);
@@ -385,8 +392,8 @@ sub list_metacpan_modules {
         $params->{sort} = [{author=>{order=>'asc'}}]  if $args{sort} eq 'author';
         $params->{sort} = [{author=>{order=>'desc'}}] if $args{sort} eq '-author';
     }
-    log_trace "MetaCPAN API query params: %s", $params;
 
+    log_trace "[MetaCPAN::Client] Requesting modules (query=%s, params=%s) ...", $query, $params;
     my $res = $mcpan->module($query, $params);
 
     _resultset_to_envres($res, $args{fields});
@@ -413,7 +420,7 @@ sub open_metacpan_module_page {
 $SPEC{open_metacpan_dist_page} = {
     v => 1.1,
     args => {
-        dist => {
+        distribution => {
             schema => 'perl::distname*',
             req => 1,
             pos => 0,
@@ -424,10 +431,199 @@ sub open_metacpan_dist_page {
     require Browser::Open;
 
     my %args = @_;
-    Browser::Open::open_browser("https://metacpan.org/release/$args{dist}");
+    Browser::Open::open_browser("https://metacpan.org/release/$args{distribution}");
     [200];
 }
 
+$SPEC{list_metacpan_distribution_versions} = {
+    v => 1.1,
+    summary => 'List all versions of a distribution',
+    description => <<'_',
+
+The versions will be sorted in a descending order.
+
+_
+    args => {
+        distribution => {
+            schema => 'perl::distname*',
+            req => 1,
+            pos => 0,
+        },
+    },
+};
+sub list_metacpan_distribution_versions {
+    my %args = @_;
+    my $res = list_metacpan_releases(
+        distribution => $args{distribution},
+        fields => ['version'],
+    );
+    return $res unless $res->[0] == 200;
+    [200, "OK", [sort { version->parse($b) <=> version->parse($a) } map {$_->{version}} @{$res->[2]}]];
+}
+
+$SPEC{download_metacpan_release} = {
+    v => 1.1,
+    summary => 'Download a release to the current directory',
+    description => <<'_',
+
+Uses <pm:HTTP::Tiny::Plugin> so you can customize download behavior using
+e.g. `HTTP_TINY_PLUGINS` environment variable.
+
+_
+    args => {
+        distribution => {
+            schema => 'perl::distname*',
+            req => 1,
+            pos => 0,
+        },
+        version => {
+            summary => 'If unspecified, will select the latest release',
+            schema => 'perl::module::release::version',
+            pos => 1,
+        },
+        overwrite => {
+            summary => 'Whether to overwrite existing downloaded file',
+            schema => 'true*',
+            cmdline_aliases => {O=>{}},
+        },
+        # XXX filename
+    },
+    examples => [
+        {
+            summary => 'Download latest release of App-orgadb distribution',
+            argv => [qw/App-orgadb/],
+            test => 0,
+            'x.doc.show_result' => 0,
+        },
+        {
+            summary => 'Download the second latest release of App-orgadb distribution',
+            argv => [qw/App-orgadb latest-1/],
+            test => 0,
+            'x.doc.show_result' => 0,
+        },
+    ],
+};
+sub download_metacpan_release {
+    my %args = @_;
+
+    my $res = list_metacpan_releases(
+        distribution => $args{distribution},
+        fields => [qw/version date author download_url/],
+    );
+    return $res unless $res->[0] == 200;
+    #use DD; dd $res;
+
+    my $rels = [sort {version->parse($b->{version}) <=> version->parse($a->{version})} @{$res->[2]}];
+    #use DD; dd $rels;
+
+    require Module::Release::Select;
+    my $rel = Module::Release::Select::select_release(
+        {detail=>1}, $args{version}, $rels);
+    #use DD; dd $rel;
+    return [404, "Version $args{version} of distribution $args{distribution} not found in releases"] unless $rel;
+
+    my $url = $rel->{download_url};
+    (my $filename = $url) =~ s!.+/!!;
+    return [412, "File '$filename' already exists, not overwriting (use -O to overwrite)"]
+        if (-f $filename) && !$args{overwrite};
+
+    open my $fh, ">", $filename
+        or return [500, "Can't open $filename for writing: $!"];
+
+    require HTTP::Tiny::Plugin;
+    log_trace "Downloading %s ...", $url;
+    my $dlres = HTTP::Tiny::Plugin->new->get($url);
+    return [500, "Can't download $url: $dlres->{status} - $dlres->{reason}"]
+        unless $dlres->{success};
+
+    print $fh $dlres->{content};
+    close $fh or return [500, "Can't write $filename: $!"];
+
+    [200, "OK", undef, {
+        'func.filename' => $filename,
+        'func.url' => $url,
+        'func.version' => $rel->{version},
+    }];
+}
+
+$SPEC{diff_metacpan_releases} = {
+    v => 1.1,
+    summary => 'Diff two release tarballs',
+    args => {
+        distribution => {
+            schema => 'perl::distname*',
+            req => 1,
+            pos => 0,
+        },
+        version1 => {
+            schema => 'perl::module::release::version',
+            req => 1,
+            pos => 1,
+        },
+        version2 => {
+            schema => 'perl::module::release::version',
+            req => 1,
+            pos => 2,
+        },
+    },
+    examples => [
+        {
+            summary => 'What changed between App-orgadb 0.014 and 0.015?',
+            argv => [qw/App-orgadb 0.014 0.015/],
+            test => 0,
+            'x.doc.show_result' => 0,
+        },
+        {
+            summary => 'What changed in the latest version of App-orgadb?',
+            argv => [qw/App-orgadb latest-1 latest/],
+            test => 0,
+            'x.doc.show_result' => 0,
+        },
+    ],
+    'x.envs' => {
+        'DIFF_METACPAN_RELEASES_DEBUG' => {
+            summary => 'Enable debugging',
+            description => <<'_',
+
+If set to true:
+- will not delete temporary directory used to store
+
+_
+            schema => 'bool*',
+        },
+    },
+};
+sub diff_metacpan_releases {
+    require File::Temp;
+    my %args = @_;
+
+    my $tempdir = File::Temp::tempdir(cleanup => !$ENV{DIFF_METACPAN_RELEASES_DEBUG});
+    log_debug "Temporary directory: %s", $tempdir;
+
+    local $CWD = $tempdir;
+
+    my $dlres1 = download_metacpan_release(
+        distribution => $args{distribution},
+        version => $args{version1},
+    );
+    return [500, "Can't download $args{distribution} version $args{version1}: $dlres1->[0] - $dlres1->[1]"]
+        unless $dlres1->[0] == 200;
+
+    my $dlres2 = download_metacpan_release(
+        distribution => $args{distribution},
+        version => $args{version2},
+    );
+    return [500, "Can't download $args{distribution} version $args{version2}: $dlres2->[0] - $dlres2->[1]"]
+        unless $dlres2->[0] == 200;
+
+    # XXX currently we just assume the two archives are tarballs
+    require App::DiffTarballs;
+    App::DiffTarballs::diff_tarballs(
+        tarball1 => $dlres1->[3]{'func.filename'},
+        tarball2 => $dlres2->[3]{'func.filename'},
+    );
+}
+
 1;
 # ABSTRACT: CLI utilities related to MetaCPAN
 
@@ -443,7 +639,7 @@ App::MetaCPANUtils - CLI utilities related to MetaCPAN
 
 =head1 VERSION
 
-This document describes version 0.006 of App::MetaCPANUtils (from Perl distribution App-MetaCPANUtils), released on 2022-02-24.
+This document describes version 0.007 of App::MetaCPANUtils (from Perl distribution App-MetaCPANUtils), released on 2023-01-15.
 
 =head1 DESCRIPTION
 
@@ -451,6 +647,12 @@ This distribution contains CLI utilities related to MetaCPAN:
 
 =over
 
+=item * L<diff-metacpan-releases>
+
+=item * L<download-metacpan-release>
+
+=item * L<list-metacpan-distribution-versions>
+
 =item * L<list-metacpan-distributions>
 
 =item * L<list-metacpan-releases>
@@ -466,6 +668,161 @@ This distribution contains CLI utilities related to MetaCPAN:
 =head1 FUNCTIONS
 
 
+=head2 diff_metacpan_releases
+
+Usage:
+
+ diff_metacpan_releases(%args) -> [$status_code, $reason, $payload, \%result_meta]
+
+Diff two release tarballs.
+
+Examples:
+
+=over
+
+=item * What changed between App-orgadb 0.014 and 0.015?:
+
+ diff_metacpan_releases(distribution => "App-orgadb", version1 => 0.014, version2 => 0.015);
+
+=item * What changed in the latest version of App-orgadb?:
+
+ diff_metacpan_releases(
+     distribution => "App-orgadb",
+   version1     => "latest-1",
+   version2     => "latest"
+ );
+
+=back
+
+This function is not exported.
+
+Arguments ('*' denotes required arguments):
+
+=over 4
+
+=item * B<distribution>* => I<perl::distname>
+
+(No description)
+
+=item * B<version1>* => I<perl::module::release::version>
+
+(No description)
+
+=item * B<version2>* => I<perl::module::release::version>
+
+(No description)
+
+
+=back
+
+Returns an enveloped result (an array).
+
+First element ($status_code) is an integer containing HTTP-like status code
+(200 means OK, 4xx caller error, 5xx function error). Second element
+($reason) is a string containing error message, or something like "OK" if status is
+200. Third element ($payload) is the actual result, but usually not present when enveloped result is an error response ($status_code is not 2xx). Fourth
+element (%result_meta) is called result metadata and is optional, a hash
+that contains extra information, much like how HTTP response headers provide additional metadata.
+
+Return value:  (any)
+
+
+
+=head2 download_metacpan_release
+
+Usage:
+
+ download_metacpan_release(%args) -> [$status_code, $reason, $payload, \%result_meta]
+
+Download a release to the current directory.
+
+Examples:
+
+=over
+
+=item * Download latest release of App-orgadb distribution:
+
+ download_metacpan_release(distribution => "App-orgadb");
+
+=item * Download the second latest release of App-orgadb distribution:
+
+ download_metacpan_release(distribution => "App-orgadb", version => "latest-1");
+
+=back
+
+Uses L<HTTP::Tiny::Plugin> so you can customize download behavior using
+e.g. C<HTTP_TINY_PLUGINS> environment variable.
+
+This function is not exported.
+
+Arguments ('*' denotes required arguments):
+
+=over 4
+
+=item * B<distribution>* => I<perl::distname>
+
+(No description)
+
+=item * B<overwrite> => I<true>
+
+Whether to overwrite existing downloaded file.
+
+=item * B<version> => I<perl::module::release::version>
+
+If unspecified, will select the latest release.
+
+
+=back
+
+Returns an enveloped result (an array).
+
+First element ($status_code) is an integer containing HTTP-like status code
+(200 means OK, 4xx caller error, 5xx function error). Second element
+($reason) is a string containing error message, or something like "OK" if status is
+200. Third element ($payload) is the actual result, but usually not present when enveloped result is an error response ($status_code is not 2xx). Fourth
+element (%result_meta) is called result metadata and is optional, a hash
+that contains extra information, much like how HTTP response headers provide additional metadata.
+
+Return value:  (any)
+
+
+
+=head2 list_metacpan_distribution_versions
+
+Usage:
+
+ list_metacpan_distribution_versions(%args) -> [$status_code, $reason, $payload, \%result_meta]
+
+List all versions of a distribution.
+
+The versions will be sorted in a descending order.
+
+This function is not exported.
+
+Arguments ('*' denotes required arguments):
+
+=over 4
+
+=item * B<distribution>* => I<perl::distname>
+
+(No description)
+
+
+=back
+
+Returns an enveloped result (an array).
+
+First element ($status_code) is an integer containing HTTP-like status code
+(200 means OK, 4xx caller error, 5xx function error). Second element
+($reason) is a string containing error message, or something like "OK" if status is
+200. Third element ($payload) is the actual result, but usually not present when enveloped result is an error response ($status_code is not 2xx). Fourth
+element (%result_meta) is called result metadata and is optional, a hash
+that contains extra information, much like how HTTP response headers provide additional metadata.
+
+Return value:  (any)
+
+
+
 =head2 list_metacpan_distributions
 
 Usage:
@@ -480,10 +837,16 @@ Arguments ('*' denotes required arguments):
 
 =item * B<author> => I<cpan::pause_id>
 
+(No description)
+
 =item * B<fields> => I<array[str]> (default: ["distribution"])
 
+(No description)
+
 =item * B<sort> => I<str> (default: "distribution")
 
+(No description)
+
 
 =back
 
@@ -514,16 +877,28 @@ Arguments ('*' denotes required arguments):
 
 =item * B<author> => I<cpan::pause_id>
 
+(No description)
+
 =item * B<fields> => I<array[str]> (default: ["module","date","author","status","maturity","version","release","abstract"])
 
+(No description)
+
 =item * B<from_date> => I<date>
 
+(No description)
+
 =item * B<sort> => I<str> (default: "module")
 
+(No description)
+
 =item * B<status> => I<str> (default: "latest")
 
+(No description)
+
 =item * B<to_date> => I<date>
 
+(No description)
+
 
 =back
 
@@ -554,24 +929,40 @@ Arguments ('*' denotes required arguments):
 
 =item * B<author> => I<cpan::pause_id>
 
+(No description)
+
 =item * B<date> => I<date>
 
 Select a single day, alternative to `from_date` + `to_date`.
 
 =item * B<distribution> => I<cpan::distname>
 
+(No description)
+
 =item * B<fields> => I<array[str]> (default: ["release","date","author","status","maturity","version","first","distribution","abstract"])
 
+(No description)
+
 =item * B<first> => I<bool>
 
+(No description)
+
 =item * B<from_date> => I<date>
 
+(No description)
+
 =item * B<sort> => I<str> (default: "-date")
 
+(No description)
+
 =item * B<status> => I<str>
 
+(No description)
+
 =item * B<to_date> => I<date>
 
+(No description)
+
 
 =back
 
@@ -602,8 +993,12 @@ Arguments ('*' denotes required arguments):
 
 =item * B<fields> => I<array[str]> (default: ["release","date","author","status","maturity","version","first","distribution","abstract"])
 
+(No description)
+
 =item * B<n> => I<posint>
 
+(No description)
+
 
 =back
 
@@ -632,7 +1027,9 @@ Arguments ('*' denotes required arguments):
 
 =over 4
 
-=item * B<dist>* => I<perl::distname>
+=item * B<distribution>* => I<perl::distname>
+
+(No description)
 
 
 =back
@@ -664,6 +1061,8 @@ Arguments ('*' denotes required arguments):
 
 =item * B<module>* => I<perl::modname>
 
+(No description)
+
 
 =back
 
@@ -721,13 +1120,14 @@ simply modify the code, then test via:
 
 If you want to build the distribution (e.g. to try to install it locally on your
 system), you can install L<Dist::Zilla>,
-L<Dist::Zilla::PluginBundle::Author::PERLANCAR>, and sometimes one or two other
-Dist::Zilla plugin and/or Pod::Weaver::Plugin. Any additional steps required
-beyond that are considered a bug and can be reported to me.
+L<Dist::Zilla::PluginBundle::Author::PERLANCAR>,
+L<Pod::Weaver::PluginBundle::Author::PERLANCAR>, and sometimes one or two other
+Dist::Zilla- and/or Pod::Weaver plugins. Any additional steps required beyond
+that are considered a bug and can be reported to me.
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2022, 2021, 2020 by perlancar <perlancar@cpan.org>.
+This software is copyright (c) 2023, 2022, 2021, 2020 by perlancar <perlancar@cpan.org>.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
